package labtest1_solution;

import java.awt.Canvas;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.RenderingHints;
import java.awt.Shape;
import java.awt.event.ComponentListener;
import java.awt.font.FontRenderContext;
import java.awt.font.LineBreakMeasurer;
import java.awt.font.TextAttribute;
import java.awt.font.TextLayout;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.awt.image.BufferStrategy;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.text.AttributedCharacterIterator;
import java.text.AttributedString;
import java.util.EventListener;
import java.util.Hashtable;
import java.util.List;
import java.util.Vector;

import javax.imageio.ImageIO;
import javax.swing.ImageIcon;

/**
 * A canvas which renders a smiley-face to the screen
 * 
 * Note: Canvas is a heavyweight superclass, which makes SmileyCanvas also
 * heavyweight. To convert this class to a lightweight, change "extends Canvas"
 * to "extends Component".
 * 
 * http://java.sun.com/products/jfc/tsc/articles/painting/src/PaintDemo.java
 * 
 * http://download.oracle.com/javase/tutorial/2d/text/drawmulstring.html
 * 
 */
public class A1Canvas extends Canvas {

	private A1ModelVariant2 a1Model;
	private LineBreakMeasurer lineMeasurer;

	// index of the first character in the paragraph.
	private int paragraphStart;

	// index of the first character after the end of the paragraph.
	private int paragraphEnd;

	public A1Canvas(A1ModelVariant2 a1Model) {
		this.a1Model = a1Model;
	}

	public Dimension getPreferredSize() {
		return new Dimension(600, 600);
	}

	/*
	 * Paint when the AWT tells us to...
	 */
	public void paint(Graphics g) {
		// Dynamically calculate size information
		// (the canvas may have been resized externally...)
		Dimension size = getSize();
		Graphics2D g2 = (Graphics2D) this.getGraphics();
		Image img = a1Model.getImage();
		img = img.getScaledInstance(size.width, size.height,
				Image.SCALE_AREA_AVERAGING);
		g2.drawImage(img, 0, 0, null);

		List<VisualMass> shapes = a1Model.getShapes();
		for (VisualMass vm : shapes) {
			g2.setColor(vm.getTheColour());
		}

		Shape textBoundingBox = a1Model.getShapes().get(0)
				.getOutlineForVisualMass(size);

		g2.setColor(Color.RED);

		drawStringAsSpecified(g2, a1Model.getTheText(), textBoundingBox);
	}

	private void drawStringAsSpecified(Graphics2D g2d,
			AttributedString theString, Shape s) {
		Rectangle2D boundingBox = s.getBounds2D();
		Point anchor = new Point((int) Math.floor(boundingBox.getX()),
				(int) Math.floor(boundingBox.getY()));
		Dimension size = new Dimension(
				(int) Math.floor(boundingBox.getWidth()),
				(int) Math.floor(boundingBox.getBounds2D().getHeight()));

		drawStringAsSpecified(g2d, theString, size, anchor);
	}

	private void drawStringAsSpecified(Graphics2D g2d,
			AttributedString theString, Dimension boundingBox, Point anchorPoint) {
		g2d.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING,
				RenderingHints.VALUE_TEXT_ANTIALIAS_LCD_HRGB);

		// body of this method derived directly from LineBreakSample.java
		// example from the Oracle website

		// Set break width to width of bounding box
		float breakWidth = boundingBox.width;

		// needed improvement: this attribute value should be manipulated by a
		// setter/getter rather than being hard-coded
		Font f = new Font("Arial", Font.PLAIN, 30);
		theString.addAttribute(TextAttribute.FONT, f);

		lineMeasurer = getLineMeasurer(theString, g2d);
		// MB: we need to derive a list of lines in advance of their rendering;
		// we need to know the total number of lines in order to derive the
		// vertical anchor of the first line
		List<TextLayout> theLines = getTextLayout(lineMeasurer, breakWidth);

		float vertSpace = theLines.get(0).getAscent()
				+ theLines.get(0).getDescent() + theLines.get(0).getLeading();
		double vertReq = theLines.size() * vertSpace;

		while (vertReq > boundingBox.height) {
			f = f.deriveFont((float) (f.getSize() - 1));
			// System.out.println(theLines.size() + "\t" + vertSpace + "\t"
			// + theLines.size() * vertSpace + "\t" + boundingBox.height
			// + "\t" + f.getSize());
			theString.addAttribute(TextAttribute.FONT, f, 0, a1Model
					.getTheString().length());
			lineMeasurer = getLineMeasurer(theString, g2d);
			theLines = getTextLayout(lineMeasurer, breakWidth);

			vertSpace = theLines.get(0).getAscent()
					+ theLines.get(0).getDescent()
					+ theLines.get(0).getLeading();
			vertReq = theLines.size() * vertSpace;
		}
		// System.out.println(theLines.size() + "\t" + vertSpace + "\t"
		// + theLines.size() * vertSpace + "\t" + boundingBox.height);

		// MB: now we start the vertical anchor at the bottom-most position and
		// move upwards according to the number of lines to be drawn. We draw
		// the lines in reverse order.
		float drawPosY = anchorPoint.y + boundingBox.height;
		for (int i = theLines.size() - 1; i >= 0; i--) {
			TextLayout layout = theLines.get(i);
			// Compute pen x position.
			float drawPosX = anchorPoint.x
					+ getDrawPosXForRightJustified(breakWidth,
							layout.getAdvance());

			// Move y-coordinate by the ascent of the layout.
			drawPosY -= layout.getAscent();

			// Draw the TextLayout at (drawPosX, drawPosY).
			layout.draw(g2d, drawPosX, drawPosY);

			// Move y-coordinate in preparation for next layout.
			drawPosY -= layout.getDescent() + layout.getLeading();
			// System.out.println(drawPosY);
		}
	}

	private LineBreakMeasurer getLineMeasurer(AttributedString theString,
			Graphics2D g2d) {
		// if (lineMeasurer == null) {
		AttributedCharacterIterator paragraph = theString.getIterator();
		paragraphStart = paragraph.getBeginIndex();
		paragraphEnd = paragraph.getEndIndex();
		// System.out.println(theString.g2d.getFont());
		FontRenderContext frc = g2d.getFontRenderContext();
		lineMeasurer = new LineBreakMeasurer(paragraph, frc);
		// Set position to the index of the first character in the
		// paragraph.
		lineMeasurer.setPosition(paragraphStart);

		return lineMeasurer;
	}

	public List<TextLayout> getTextLayout(LineBreakMeasurer lineMeasurer,
			float breakWidth) {
		List<TextLayout> theLines = new Vector<TextLayout>();
		// Get lines until the entire paragraph has been displayed.
		while (lineMeasurer.getPosition() < paragraphEnd) {
			// Retrieve next layout. A cleverer program would also cache
			// these layouts until the component is re-sized.
			TextLayout layout = lineMeasurer.nextLayout(breakWidth);
			// System.out.println(layout);// layout.
			theLines.add(layout);
		}
		return theLines;
	}

	public float getDrawPosXForLeftJustified() {
		return 0;
	}

	public float getDrawPosXForRightJustified(float widthAvailable,
			float lengthOfLine) {
		return widthAvailable - lengthOfLine;
	}

}
